from typing import Iterable, TypeVar

_T = TypeVar("_T")


class OrderedSet(set):
    def __init__(self, iterable=None):
        self._data = {}
        if iterable is not None:
            for item in iterable:
                self.add(item)

    def add(self, element: _T):
        self._data[element] = None

    def discard(self, element: _T):
        if element in self._data:
            del self._data[element]

    def __contains__(self, item):
        return item in self._data

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._data.keys())

    def __repr__(self):
        return f"PartialSet({list(self._data.keys())})"

    def __eq__(self, other):
        if isinstance(other, OrderedSet):
            return self._data == other._data
        return False

    def __ne__(self, value: object) -> bool:
        return not self == value

    def clear(self):
        self._data.clear()

    def pop(self):
        if not self._data:
            raise KeyError("pop from an empty OrderedSet")
        key = next(iter(self._data))
        self.discard(key)
        return key

    def update(self, s: Iterable[_T]) -> None:
        for item in s:
            self.add(item)

    def __getitem__(self, index):
        return list(self._data.keys())[index]

    def __reversed__(self):
        return reversed(list(self._data.keys()))

    def copy(self):
        return OrderedSet(self._data.keys())



from typing import Any, Optional

from pydantic import BaseModel, Field


class StopEvent(BaseModel):
    stopped_by: Optional[str] = Field(default=None, description="A note about who stopped.")
    stopped_reason: Optional[str] = Field(default=None, description="A note about why stopped.")
    additional_info: Optional[dict[str, Any]] = Field(
        default_factory=dict, description="Additional information."
    )



import collections.abc as c
import typing as t
from inspect import iscoroutinefunction
from typing import Optional

from blinker.base import Signal as BlinkerSignal
from pydantic import BaseModel, Field

from pi_blink.events import StopEvent
from pi_blink.utils.id_utils import get_full_class_name


class SignalReturn(BaseModel):
    listeners: list[t.Any] = Field(
        default_factory=list, description="The listeners that were called."
    )
    results: list[t.Any] = Field(
        default_factory=list, description="The results from the listeners."
    )
    stopped: bool = Field(default=False, description="Whether the signal was stopped.")
    stop_event: Optional[StopEvent] = Field(
        default=None, description="The stop event that stopped the signal."
    )


class Signal(BlinkerSignal):

    def extended_send(
        self,
        sender: t.Any | None = None,
        /,
        *,
        _async_wrapper: (
            c.Callable[[c.Callable[..., c.Coroutine[t.Any, t.Any, t.Any]]], c.Callable[..., t.Any]]
            | None
        ) = None,
        **kwargs: t.Any,
    ) -> SignalReturn:
        """Extends blinker.send to support cancelling receivers.\n
        Returns has also been modified to return a SignalReturn object.\n

        \n{0}""".format(
            BlinkerSignal.send.__doc__
        )

        sr = SignalReturn()
        if self.is_muted:
            return sr

        for receiver in self.receivers_for(sender):

            if iscoroutinefunction(receiver):
                if _async_wrapper is None:
                    raise RuntimeError("Cannot send to a coroutine function.")

                result = _async_wrapper(receiver)(sender, **kwargs)
            else:
                result = receiver(sender, **kwargs)
            if isinstance(result, StopEvent):
                sr.stopped = True
                sr.stop_event = result
                result.stopped_by = get_full_class_name(receiver)
                return sr
            sr.listeners.append(receiver)
            sr.results.append(result)

        return sr

    async def extended_send_async(
        self,
        sender: t.Any | None = None,
        include_stop_event: bool = False,
        /,
        *,
        _sync_wrapper: (
            c.Callable[[c.Callable[..., t.Any]], c.Callable[..., c.Coroutine[t.Any, t.Any, t.Any]]]
            | None
        ) = None,
        **kwargs: t.Any,
    ) -> SignalReturn:
        """Extends blinker.send_async to support cancelling receivers.\n
        Returns has also been modified to return a SignalReturn object.\n
        
        \n{0}""".format(
            BlinkerSignal.send_async.__doc__
        )

        sr = SignalReturn()
        if self.is_muted:
            return sr

        for receiver in self.receivers_for(sender):
            if not iscoroutinefunction(receiver):
                if _sync_wrapper is None:
                    raise RuntimeError("Cannot send to a non-coroutine function.")

                result = await _sync_wrapper(receiver)(sender, **kwargs)
            else:
                result = await receiver(sender, **kwargs)
            if isinstance(result, StopEvent):
                sr.stopped = True
                sr.stop_event = result
                result.stopped_by = get_full_class_name(receiver)
                return sr
            sr.listeners.append(receiver)
            sr.results.append(result)

        return sr




def get_full_class_name(str_cls_or_obj: str | type | object) -> str:
    """Get the full class name of a class or object.
    Args:
        str_cls_or_obj (str | type | object): The string or class or object.
    Returns:
        str: The full class name.
    """
    if isinstance(str_cls_or_obj, str):
        return str_cls_or_obj
    cls = str_cls_or_obj if isinstance(str_cls_or_obj, type) else str_cls_or_obj.__class__
    module = cls.__module__
    name = cls.__qualname__
    if module is None or module == str.__class__.__module__:
        return name  # Avoid reporting __builtin__
    return module + "." + name


import logging
import multiprocessing
import threading
from collections import defaultdict
from dataclasses import dataclass, field
from enum import IntEnum
from functools import wraps
from inspect import iscoroutinefunction
from multiprocessing import Lock
from typing import Any, Callable, Optional

from blinker._utilities import make_id
from blinker.base import ANY, ANY_ID

from pi_blink.blinker.base import Signal
from pi_blink.ordered_set import OrderedSet
from pi_blink.utils.id_utils import get_full_class_name

log = logging.getLogger(__name__)


class EventPriority(IntEnum):
    """Enum class for event priorities.
    Initial will be the first to be called, Monitor will be the last.
    Monitor is used for listeners that need to monitor the event
    and shouldn't change the event.
    """

    INITIAL = 0
    EARLIEST = 10
    EARLIER = 20
    EARLY = 30
    NORMAL = 40
    LATE = 50
    LATER = 60
    LATEST = 70
    LAST = 80
    MONITOR = 90


# Define a class for the event
@dataclass
class PriorityListener:
    """Listener function with priority."""

    listener: Callable
    priority: EventPriority = field(default_factory=lambda: EventPriority.NORMAL)

    def __post_init__(self):
        if self.priority is None:
            self.priority = EventPriority.NORMAL

    def __lt__(self, other):
        return self.priority < other.priority

    def __gt__(self, other):
        return self.priority > other.priority

    def __le__(self, other):
        return self.priority <= other.priority

    def __ge__(self, other):
        return self.priority >= other.priority

    def __ne__(self, other):
        return self.priority != other.priority

    def __eq__(self, other):
        return self.priority == other.priority

    def __hash__(self):
        return hash(self.priority)

    def __repr__(self):
        return f"PL({self.listener} -[{self.priority.name} {self.priority.value}])"


class ListenerRegistry:
    """Singleton class for the listener registry."""

    _instance = None
    _lock = Lock()

    def __new__(cls):
        with cls._lock:
            if cls._instance is None:
                cls._instance = super(ListenerRegistry, cls).__new__(cls)
                cls._instance._initialize()
            return cls._instance

    def _initialize(self) -> None:
        self.receiver_priorities: dict[Any, dict[Any, EventPriority]] = defaultdict(dict)
        self._signals: dict[str, Signal] = {}

    def register_listener(
        self,
        signal_class: type,
        func: Callable,
        sender: Any,
        weak: bool = True,
        priority: Optional[EventPriority] = None,
    ):
        """
        Register a listener for a signal class.
        """
        sn = get_full_class_name(signal_class)
        fn = func.__module__ + "." + func.__name__
        priority = priority or EventPriority.NORMAL
        fid = make_id(func)

        ## sender_id mimics base.py
        sender_id = ANY_ID if sender is ANY else make_id(sender)
        rps = self.receiver_priorities[sn]
        rps[fid] = priority

        if sn not in self._signals:
            self._signals[sn] = Signal(sn)

        sig = self._signals[sn]

        sig.connect(func, sender=sender, weak=weak)

        ## Reorder the receivers by priority
        new_by_sender: dict[Any, set[Any]] = defaultdict(OrderedSet)
        for sender_id, receiver_set in sig._by_sender.items():
            ## sort the new receiver set by priority
            d = {k: rps.get(k, EventPriority.NORMAL) for k in receiver_set}
            sorted_items = sorted(d.items(), key=lambda i: i[1])
            new_by_sender[sender_id] = OrderedSet([k for k, _ in sorted_items])
        sig._by_sender = new_by_sender

        log.debug(
            f"Registering listener: {func} {fn} {sn} thread: {threading.get_ident()} nsig: {len(self._signals)}, process_id: {multiprocessing.current_process()}"
        )

    def send_event(self, event: Any, sender: Any = None):
        """
        Send an event to all listeners of the event.
        Args:
            event (Any): The event object.
            sender (Any): The sender of the event.
        """
        log.debug(
            f"Sending event: {event} {sender}: thread: {threading.get_ident()} nsig: {len(self._signals)}, process_id: {multiprocessing.current_process()}"
        )
        sig = self._signals.get(get_full_class_name(event))
        if not sig:
            raise ValueError(
                f"No signal found for event: {get_full_class_name(event)}, sender: {sender}"
            )
        return sig.extended_send(event)

    async def asend_event(self, event: Any, sender: Any = None):
        """
        Send an event to all listeners of the event asynchronously.
        Args:
            event (Any): The event to send.
            sender (Any): The sender of the event.

        """
        log.debug(
            f"Async sending event: {event} {sender}: thread: {threading.get_ident()} nsig: {len(self._signals)}, process_id: {multiprocessing.current_process()}"
        )

        sig = self._signals.get(get_full_class_name(event))
        if not sig:
            raise ValueError(
                f"No signal found for event: {get_full_class_name(event)}, sender: {sender}"
            )
        await sig.extended_send_async(event)

    def change_priority(self, signal_class, func, new_priority: EventPriority):
        """Change the priority of a listener.
        Args:
            signal_class (type): The signal class.
            func (Callable): The listener function.
            new_priority (EventPriority): The new priority.

        """
        sn = get_full_class_name(signal_class)
        rps = self.receiver_priorities[sn]
        fid = make_id(func)
        old_priority = rps.get(fid, None)
        if old_priority == new_priority:
            return
        rps[fid] = new_priority
        sig = self._signals[sn]
        new_by_sender: dict[Any, set[Any]] = defaultdict(OrderedSet)
        for sender_id, receiver_set in sig._by_sender.items():
            ## sort the new receiver set by priority
            d = {k: rps.get(k, EventPriority.NORMAL) for k in receiver_set}
            sorted_items = sorted(d.items(), key=lambda i: i[1])
            new_by_sender[sender_id] = OrderedSet([k for k, _ in sorted_items])
        sig._by_sender = new_by_sender

    def delete_listener(self, signal_class, func):
        """Delete a listener."""
        sn = get_full_class_name(signal_class)
        sig = self._signals.get(sn)
        if sig:
            sig.disconnect(func)
        else:
            raise ValueError(f"No signal found for event: {sn}")


# Singleton instance of the registry
_listener_registry = ListenerRegistry()


def listener(signal_class: Any, sender: Any = ANY, priority: Optional[EventPriority] = None):
    """
    Decorator for registering a listener.
    Args:
        signal_class (type): The signal class.
        sender (Any): The sender of the signal.
        priority (EventPriority): The priority of the listener.
    Example:
        class MyEvent:
            pass
        @listener(MyEvent)
        def my_func(event : MyEvent):
            pass
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        _listener_registry.register_listener(
            signal_class, sender=sender, func=wrapper, priority=priority
        )
        return wrapper

    return decorator


def alistener(signal_class: Any, sender: Any = ANY, priority: EventPriority | None = None):
    """
    Decorator for registering an async listener.
    Args:
        signal_class (type): The signal class.
        sender (Any): The sender of the signal.
        priority (EventPriority): The priority of the listener.
    Example:
        class MyEvent:
            pass
        @alistener(MyEvent)
        async def my_func(event : MyEvent):
            pass
    """

    def decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            return await func(*args, **kwargs)

        ## check if is coroutine
        if not iscoroutinefunction(func):
            raise ValueError(
                f"Function {func} is not a coroutine.\n"
                f"Use @listener instead of @alistener, or async def {func}."
            )
        _listener_registry.register_listener(
            signal_class, sender=sender, func=wrapper, priority=priority
        )
        return wrapper

    return decorator


class Blink:
    """
    Class to interface with listeners and events.
    """

    @staticmethod
    def send(event: Any, sender: Any = None):
        """Send an event.
        Args:
            event (Any): The event to send.
            sender (Any): The sender of the event.
        """
        return _listener_registry.send_event(event, sender)

    @staticmethod
    async def asend(event: Any, sender: Any = None):
        """Send an event asynchronously.
        Args:
            event (Any): The event to send.
            sender (Any): The sender of the event.
        """
        return await _listener_registry.asend_event(event, sender)

    @staticmethod
    def change_listener_priority(event: Any, func: Callable, new_priority: EventPriority):
        """Change the priority of a listener.
        Args:
            event (Any): The event class.
            func (Callable): The listener function.
            new_priority (EventPriority): The new priority."""
        _listener_registry.change_priority(event, func, new_priority)

    @staticmethod
    def delete_listener(event: Any, func: Callable):
        """Delete a listener.
        Args:
            event (Any): The event class.
            func (Callable): The listener function.
        """
        _listener_registry.delete_listener(event, func)

    @staticmethod
    def listener(signal_class, sender: Any = ANY, priority: EventPriority | None = None):
        """Register a listener.
        Args:
            signal_class (type): The signal class.
            sender (Any): The sender of the signal.
            priority (EventPriority): The priority of the listener.
        Returns:
            Callable: The listener function.
        """
        return listener(signal_class=signal_class, sender=sender, priority=priority)

    @staticmethod
    def alistener(signal_class: Any, sender: Any = ANY, priority: EventPriority | None = None):
        """Register an async listener.
        Args:
            signal_class (type): The signal class.
            sender (Any): The sender of the signal.
            priority (EventPriority): The priority of the listener.
        Returns:
            Callable: The listener function.
        """
        return alistener(signal_class=signal_class, sender=sender, priority=priority)


blink = Blink  # Alias for Blink class
